import abc
from typing import List

from idp_authentication.users.domain.entities.user import User


class RepositoryPort(abc.ABC):
    @abc.abstractmethod
    def get_or_none(self, **kwargs) -> User:
        """Get a record by its attributes."""

    @abc.abstractmethod
    def get_first(self, **filters):
        """Get the first record, after applying filters"""

    @abc.abstractmethod
    def all(self):
        """Get all records"""

    @abc.abstractmethod
    def create(self, **kwargs):
        """Create a new record."""

    @abc.abstractmethod
    def update_record(self, record, **kwargs):
        """Update a record."""


class UserRepositoryPort(RepositoryPort, abc.ABC):
    """User repository port."""

    @abc.abstractmethod
    def get_users_with_access_to_records(self, *args, **kwargs) -> List[User]:
        """Get users with access to records."""

    @abc.abstractmethod
    def remove_user_role(self, user, role):
        """Remove user role."""


class UserRoleRepositoryPort(RepositoryPort, abc.ABC):
    """User role repository port."""

    @abc.abstractmethod
    def delete(self, record):
        """Delete a record."""



from sqlalchemy import or_

from idp_authentication.users.base_classes.base_repository import BaseRepository
from idp_authentication.users.domain.entities.user import User
from idp_authentication.users.domain.entities.user_role import UserRole
from idp_authentication.users.domain.ports.repository import UserRepositoryPort


class UserRepository(UserRepositoryPort, BaseRepository):
    entity = User

    def get_users_with_access_to_records(self, *args, **kwargs):
        # Todo: Implement this method
        """Get users with access to records."""
        user_roles = kwargs.get("roles")
        # app_entity_type = kwargs.get("app_entity_type")
        # record_identifier = kwargs.get("record_identifier")
        roles = (
            self.session.query(UserRole)
            .join(User)
            .filter(
                UserRole.role.in_(user_roles),
                or_(
                    UserRole.app_entities_restrictions is not None,  # noqa: E711
                ),
            )
            .all()
        )
        role_ids = [role.id for role in roles]
        return (
            self.session.query(self.entity)
            .join(UserRole)
            .filter(
                self.entity.is_active.is_(True),
                User.user_roles.any(UserRole.id.in_(role_ids)),
            )
            .all()
        )

    def remove_user_role(self, user: User, role: UserRole):
        user.user_roles.remove(role)



from dataclasses import dataclass, field
from datetime import date, datetime
from typing import Optional

from idp_authentication.users.base_classes.base_entity import BaseEntity
from idp_authentication.users.domain.entities.user_role import UserRole


@dataclass(kw_only=True)
class User(BaseEntity):
    username: str = field(default_factory=str)
    email: str = field(default_factory=str)
    first_name: str = field(default_factory=str)
    last_name: str = field(default_factory=str)
    is_active: bool = field(default_factory=bool)
    is_superuser: bool = field(default_factory=bool)
    is_staff: bool = field(default_factory=bool)
    last_login: Optional[datetime] = field(default=None)
    date_joined: Optional[date] = field(default=None)
    user_roles: Optional[list[UserRole]] = field(default_factory=list)

    def get_full_name(self):
        """
        Returns the first_name plus the last_name, with a space in between.
        """
        full_name = f"{self.first_name} {self.last_name}"
        return full_name.strip()

    def get_short_name(self):
        """Returns the short name for the user."""
        return self.first_name



from typing import Union

from idp_authentication.custom_types import UserTenantData
from idp_authentication.users.base_classes.base_use_case import UseCasePort
from idp_authentication.users.domain.entities.user import User
from idp_authentication.users.domain.ports.unit_of_work import UsersUnitOfWorkPort

user_data_keys = [
    "username",
    "email",
    "first_name",
    "last_name",
    "is_active",
    "is_staff",
    "is_superuser",
    "date_joined",
]


class CreateOrUpdateUserUseCase(UseCasePort):
    def __init__(self, users_unit_of_work: UsersUnitOfWorkPort):
        self.user_tenant_data = None
        self.users_unit_of_work = users_unit_of_work

    def keep_keys(self, keys: list) -> UserTenantData:
        return {k: v for k, v in self.user_tenant_data.items() if k in keys}

    def execute(self, data: UserTenantData) -> User:
        self.user_tenant_data = data
        user_data = self.keep_keys(user_data_keys)

        with self.users_unit_of_work as uow:
            user = uow.user_repository.get_or_none(
                username=self.user_tenant_data.get("username")
            )

            if user:
                user = uow.user_repository.update_record(user, **user_data)
            else:
                user = uow.user_repository.create(**user_data)

            uow.commit()
            return user



from idp_authentication.conftest import (
    TEST_PERMISSION,
    TEST_ROLE_1,
    VEHICLE_APP_ENTITY_IDENTIFIER,
)


def test_get_allowed_app_entity_records_with_permission(container, user):
    get_allowed_app_entity_records = (
        container.users_module().use_cases().get_allowed_record_identifiers_use_case()
    )
    records = get_allowed_app_entity_records.execute(
        user=user,
        role=TEST_ROLE_1,
        app_entity_type=VEHICLE_APP_ENTITY_IDENTIFIER,
        permission=TEST_PERMISSION,
    )
    assert len(records) > 0


def test_get_allowed_app_entity_records_with_no_permission(container, user_2):
    get_allowed_app_entity_records = (
        container.users_module().use_cases().get_allowed_record_identifiers_use_case()
    )
    records = get_allowed_app_entity_records.execute(
        user=user_2,
        role=TEST_ROLE_1,
        app_entity_type=VEHICLE_APP_ENTITY_IDENTIFIER,
        permission=None,
    )
    assert len(records) == 0



from typing import Any, Optional, TypedDict, TypeVar, Union

T = TypeVar("T")
Role = str
ALL = "all"

AppEntityType = str
AppEntityIdentifier = str
PermissionIdentifier = str

# AppEntitiesRestrictionsDict = {<entity_type>: [1, 2]}
AppEntitiesRestrictionsDict = dict[AppEntityType, list[AppEntityIdentifier]]

# PermissionRestrictionsDict = {"perform_operation_1": {<entity_type>: [1, 2]}, "perform_operation_2": false}
PermissionRestrictionsDict = dict[
    PermissionIdentifier, dict[AppEntityType, list[AppEntityIdentifier] | bool]
]


class AppSpecificConfigs(TypedDict):
    app_entities_restrictions: Optional[dict[str, list]]
    permission_restrictions: dict[str, Union[bool, Any]]


UserAppSpecificConfigs = dict[Role, AppSpecificConfigs]


class UserTenantData(TypedDict):
    first_name: str
    last_name: str
    username: str
    email: str
    is_active: bool
    is_staff: bool
    is_superuser: bool
    date_joined: str
    app_specific_configs: UserAppSpecificConfigs


class AppEntityEventMessage(TypedDict):
    app_identifier: str
    app_entity_type: str
    label: str
    record_identifier: Any
    deleted: bool


AppIdentifier = str
TenantIdentifier = str


UserRecordAppSpecificConfigs = dict[
    AppIdentifier, dict[TenantIdentifier, AppSpecificConfigs]
]


class UserRecordDict(TypedDict):
    first_name: str
    last_name: str
    username: str
    email: str
    is_active: bool
    is_staff: bool
    is_superuser: bool
    date_joined: str
    app_specific_configs: UserRecordAppSpecificConfigs


class AppEntityRecordEventDict(TypedDict):
    app_identifier: str
    app_entity_type: str
    record_identifier: Any
    deleted: bool
    label: Optional[str]



from sqlalchemy import Column, ForeignKey, String

from idp_authentication.users.infrastructure.database.sqlalchemy.base import (
    Base,
    BaseModelMixin,
)
from idp_authentication.users.infrastructure.database.sqlalchemy.custom_orm_column_types import (
    UUID,
    JsonField,
)


class UserRoleModel(Base, BaseModelMixin):
    __tablename__ = "user_roles"
    __table_args__ = {"extend_existing": True}

    user_id = Column(UUID(), ForeignKey("users.id"))
    role = Column(String(40), nullable=False)
    app_entities_restrictions = Column(JsonField(), nullable=False)
    permission_restrictions = Column(JsonField(), nullable=False)



from dataclasses import dataclass


@dataclass
class ExceptionBase(Exception):
    """Base class for all exceptions in this module."""

    message: str

    def __str__(self):
        return self.message


@dataclass
class AccessDenied(ExceptionBase):
    """Raised when a user is not allowed to access a resource."""

    message: str = "Access denied"


@dataclass
class UnknownAppEntityType(ExceptionBase):
    """Raised when an unknown app entity type is requested."""

    def __init__(self, app_entity_type: str):
        self.message = f"Unknown app entity: {app_entity_type}!"


@dataclass
class UserNotFound(ExceptionBase):
    """Raised when a user is not found."""

    def __init__(self, username: str):
        self.message = f"User with username={username} not found!"



from idp_authentication.conftest import TEST_ROLE_1, VEHICLE_APP_ENTITY_IDENTIFIER


def test_user_with_access_to_app_entity_record(container, vehicle, user):
    user_with_access_to_app = (
        container.users_module()
        .use_cases()
        .get_users_with_access_to_app_entity_record_use_case()
    )
    user_with_access_to_app.execute(
        app_entity_type=VEHICLE_APP_ENTITY_IDENTIFIER,
        record_identifier=[2],
        roles=[TEST_ROLE_1],
    )
